#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>

int check_num(char *para)
{
  // check the input number is integer.
  int i, len;
  len = strlen(para);
  for ( i = 0 ; i < len ; i++ )
  { 
    if ( para[i] < '0' || para [i] > '9' )
 	return 1;		
  }
  if ( atoi(para) == 0 )
 	return 1;		
  return 0; 
}

void fips140(unsigned char * buf)
{ 
   int i, j, one_bits_count = 0, zero_bits_count=0;


   printf("FIPS 140-1 Test Result\n");


// Monobit Test    
 
   printf("### Monobit Test (Passed if between 9654 and 10346)\n");
   for ( i = 0 ; i < 2500 ; i ++) {
  	for ( j = 0 ; j < 8 ; j++ ) {
	   if ( (buf[i] & (0x01<<j)) != 0 )
                one_bits_count++ ;
	   else
                zero_bits_count++ ;
	}
   } 
   if ( one_bits_count > 10346  || one_bits_count < 9654 )
       printf("    FAIL! There are %d bit 1 and %d bit 0!\n\n", one_bits_count, zero_bits_count);
   else
       printf("    PASS! There are %d bit 1 and %d bit 0!\n\n", one_bits_count, zero_bits_count);

//Poker Test

   printf("### Poker Test (Passed if between 1.03 and 57.4)\n");
   unsigned char L4bit[2500], H4bit[2500];
   unsigned int poker_counts[16] = {0};
   unsigned int total4bit = 0 ;
   float poker_value = 0.0;

   for ( i = 0 ; i < 2500 ; i++ ) {

      L4bit[i] = buf[i] & 0xF ;
      H4bit[i] = ( buf[i] & 0xF0 ) >> 4 ;

      poker_counts[ L4bit[i] ]++ ;
      poker_counts[ H4bit[i] ]++ ;
   }
   
   for ( i = 0 ; i < 16 ; i ++ ) {
       total4bit += poker_counts[i] *  poker_counts[i];
   }
   
   poker_value = ( ((16.0 / 5000.0) * (float)total4bit ) - 5000.0 );
 
  
   if ( poker_value < 1.03 || poker_value > 57.4 )
       printf("    FAIL!  Poker value = %3f\n\n", poker_value);
   else
       printf("    PASS!  Poker value = %3f\n\n", poker_value);

// Runs Test

   printf("### Runs Test\n");
   unsigned char RunsArray[20000]={'0'}, *flag0, *flag1;
   int zeros[6] = {0} , ones[6] = {0} ;  
   int off = 0, count0 = 1, count1 = 1, check_34bits = 0, check_runs = 1 ;

   for ( i = 0 ; i < 2500 ; i++ ) {
       for ( j = 0 ; j < 8 ; j++ )  {
	   if ( ( buf[i] & ( 0x01 << j ) ) != 0 )
               RunsArray[ j + off ] = '1';
 	   else
               RunsArray[ j + off ] = '0';
       }
       off += 8 ;
   }
   
   for ( i = 0 ; i < 19999 ; i++ ) {
       if ( RunsArray[i] == '0' ) {
          if ( RunsArray[i] == RunsArray[i+1] )
              count0 ++ ;
          else {
              if ( count0 < 6 )
                  zeros[count0 - 1]++ ;
              else {
                  zeros[5]++ ;
                  if ( count0 > 34 ) 
                      check_34bits = 1;
	      }
              count0 = 1;
          }
       }
       else { 
           if ( RunsArray[i] == RunsArray[i+1] )
              count1 ++ ;
           else {
              if ( count1 < 6 )
                  ones[count1 - 1]++ ;
              else {
                  ones[5]++ ;
                  if ( count1 > 34 ) 
                      check_34bits = 1;
	      }
              count1 = 1;
          }
       }
   }
   
   if ( RunsArray[19999] == '0' ) {
        if ( count0 == 1 ) 
            zeros[0]++ ;
        else {
               if ( count0 < 6 )
                  zeros[count0 - 1]++ ;
               else {
                  zeros[5]++ ;
                  if ( count0 > 34 ) 
                      check_34bits = 1;
               }
        }  
   }
   else {
        if ( count1 == 1 ) 
            ones[0]++ ;
        else {
               if ( count1 < 6 )
                  ones[count1 - 1]++ ;
               else {
                  ones[5]++ ;
                  if ( count1 > 34 ) 
                      check_34bits = 1;
               }
        }  
   }

   int interval[6][2] = { {2267, 2733}, {1079, 1421}, {502, 748}, {223, 402}, {90, 223}, {90,223} } ;
   char  str[6][30] = { "(Required Interval 2267-2733)", 
                        "(Required Interval 1079-1421)", 
                        "(Required Interval 502-748)",
                        "(Required Interval 223-402)", 
                        "(Required Interval 90-223)", 
                        "(Required Interval 90-223)" }; 

   printf("   %6s%8s | %6s%8s\n", "Ones", "Result", "Zeros","Result");

   for ( i = 0 ; i < 6 ; i++ ) { 
      if ( ones[i] > interval[i][0] && ones[i] < interval[i][1]) 
           flag1 = "PASS";
      else
           flag1 = "FAIL";

      if ( zeros[i] > interval[i][0] && zeros[i] < interval[i][1] )
           flag0 = "PASS";
      else
           flag0 = "FAIL";
 
      if ( !strcmp(flag0,"FAIL") || !strcmp(flag1, "FAIL") )
          check_runs = 0;

      if ( i < 5 )  
         printf ("%2d:%6d%8s | %6d%8s  %-40s\n", i+1, ones[i], flag1, zeros[i], flag0, str[i]);
      else 
         printf ("6+:%6d%8s | %6d%8s  %-40s\n", ones[i], flag1, zeros[i], flag0, str[i]);
   }

   if ( check_runs )
       printf("   Test Result: PASS!\n\n");
   else
       printf("   Test Result: FAIL!\n\n");

// Long Run Test

   printf("### Long Run Test\n");
   if ( check_34bits ) 
      printf ("    FAIL! A run of length 34 or more of either 0s or 1s exists.\n");
   else
      printf ("    PASS! No a run of length 34 or more of either 0s or 1s.\n");

}


int main (int argc, char* argv[])
{
 int HW_RNG ;
 FILE  * fp = NULL;
 unsigned long total, volume = 1024; 
 int maxvolume;
 int i, check_flag = 1, count = 0, linelen = 64 , check_fips = 0;
 char *para;
 char *usage="USAGE: rngtest [-b number] [-n number] [-s] [-f]\n"
	     "Option:\n"
	     "       -b: Specify how many bytes you want to generate.\n"
	     "           Default is 1024. Zero is unacceptable.\n"
	     "           The maximum is 2,147,483,647.\n"
	     "       -n: Specify how many hexadecimals per line.\n"
	     "           Default is 64. Zero is unacceptable.\n" 
	     "       -s: Silent mode. Do not display any hexadecimal.\n" 
	     "       -f: FIPS 140-1 test. If you run this, other options will be disabled.\n"; 

 // Check the program with parameters or not.   
 if ( argc >= 2 )
  {
        argv++;
	count++;
	while ( *argv )
        {
          para = *argv++;
	  count++ ;
          while ( *para )
	  {
	     if ( para[0] != '-' || strlen(para) != 2)
	     {
		printf("%s",usage);
		exit(0);
	     }
	     else
	     {
               // Check the parameter
               switch ( para[1] )
	       {
                case 'b':
			  if ( count > argc )
			    {
		                printf("%s",usage);
				exit(0);				  
			    } 				  
			  volume = atoi(*argv);
                          i = strlen(*argv) ;
                          
                          // Check the number is over than the limitation.
                          if ( i > 10 || ( i == 10 && strcmp(*argv, "2147483647") > 0 ) )
                            {
                              printf("Error:\n");
                              printf("  The number is greater than 2,147,483,647.\n");
                              printf("  Please re-input again!\n");
                              exit(1);
                            }
  			  if ( check_num(*argv++) )
			    {
 				printf("%s",usage);
				exit(0);
			    }
			  break;
               case 'n':
			  if ( count > argc )
			    {
		                printf("%s",usage);
				exit(0);				  
			    }				  
			  linelen = atoi(*argv);
  			  if ( check_num(*argv++) )
			    {
 				printf("%s",usage);
				exit(0);
			    }
                          break;
                case 'f':
                          check_fips = 1;
                          break;
                case 's':
                          check_flag = 0;
                          break;
                default:
		     	  printf("%s",usage);
			  exit(0);
                          break;
  	       }
	       count++;
	     }
  	     break;					  
	  }
        }
	
  }
 
 int loop = 1;
 i = 1;
 total = volume;

 // Check to run FIPS 140-1 test or not
  if ( check_fips ) {
     volume = 2500 ;
     maxvolume = 2500 ;
     loop = 1;
 }
 else { 
   //Check the input bytes.
    if ( volume > 8193 ) {
          maxvolume = 8192; 
          loop = (volume / maxvolume) + 1;
          if ( (volume % 8192) == 0 )
               loop --;
     }
     else 
          maxvolume = volume;
 }

 // Open the HW_RNG device
 HW_RNG = open ("/dev/hwrandom", O_RDONLY);
 if ( HW_RNG <= 0 )
 {
   printf("Error: Device not ready!!\n");
   exit(0);
 }

 if (!check_fips) 
     printf("Start to generate %lu bytes.\n", volume);

 while (loop) {
     unsigned char buf[maxvolume];
     int size;
     size_t r, off;
     size = sizeof(buf);
     off = 0 ;
      
     // Read the data from the HW/RNG
     r = read(HW_RNG, buf + off , size) ;
     size -= r ;

     if (size)
       {
         printf("Error: Read error!\n");
         exit(1);
       }
// Run FIPS 140-1 test or not
    if (check_fips) 
        fips140( buf );
    else {
        // Display hexadecimal or not
        if (check_flag) {
  	    for ( i = 0; i < maxvolume; i++) {
                if ( linelen == 1 )
                    linelen = 2 ;

                if ( (i % (linelen / 2) ) == 0 )
                    printf("\n") ;

                if ( buf[i] > 0xF )
                    printf("%2X", buf[i]);
                else 
                    printf("0%X", buf[i]);
             }
             printf("\n");
        }
        else
             printf("%10d. %d bytes generated.\n", i, maxvolume);
   
        volume -= maxvolume;

        if ( volume < 8193 )
            maxvolume = volume;
    }

    i++;     
    loop-- ; 
 }

 if (!check_fips) 
     printf("Finished. %lu bytes generated.\n", total);

 close (HW_RNG); 
 return 0;
}
